/* Nils Olav Selåsdal <NOS@Utel.no> made most this mess.
   
   See RFC-2229 for the protocol.
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>

enum { LINESZ = 6144 };

enum Command {Define,Match,ShowDB,ShowStrats};
static Biobuf in,out;
static uchar ibuf[LINESZ];
static int fd;
static char *dictdb = "*";
static char *strategy = "substring";
static char *progname;
void usage(void);

void
could_not(char *what,char *r)
{
	fprint(2,"Error: Could not %s\n",what);
	if(r) {
		while(*r && isdigit(*r++))
			;
		print("%s\n",r);
	}
	close(fd);
	exits("failed");
}

char *
dict_line(void)
{
	char *r;
	int l;

	r = Brdline(&in,'\n');
	if(r == nil)
		could_not("get response",nil);

	l = Blinelen(&in) - 1;
	if(l > 0 && r[l-1] == '\r')
		r[l-1] = 0;
	else {
		fprint(2,"Error, bad line, possible too long");
		r[l] = 0;
	}

	return r;	
}
	
char *
dict_command(char *cmd)
{
	Bprint(&out,"%s\r\n",cmd);
	Bflush(&out);

	return  dict_line();
}

void
shake_hands(void)
{
	char *r;

	//slurp initial greeting
	r = dict_line(); 
	if(atoi(r) != 220) {
		char msg[255];
		snprint(msg,sizeof(msg),"Connect: %s",r);
		could_not(msg,r);
	}
	//identify ourself
	dict_command("CLIENT plan9/define"); //I give a rats if the server doesn't swallow it.	
}

void 
read_definitions(int count,char *what)
{
	int state = 0; //0 slurp 151's , 1 slurp definition

	while(count+1) {
		char *r;
		
		r = dict_line();

		if(state == 0) {
			switch(atoi(r)) {
			case 250: // We really shouln't get thisone here..
				return; //thats all folks
			case 151: {
				char *args[2] = {nil,nil};
				int l = strlen(what) + 4;
				state = 1;
				--count;
				tokenize(l >= strlen(r) ? r:r+l,args,2);
				print("From %s:\n",args[1]);
				break;
				}
			default:
				could_not("get all definitions",r);
				break;
			}
		} else if (state == 1) { //duh
			if(strcmp(".",r) == 0) {
				print("\n"); 
				state = 0;
			} else 
				print("%s\n",r);
		}
	}
}

int
need_quote(char *p)
{
	Rune r;

	while(*p) {
		p += chartorune(&r,p);
		if(isspacerune(r))
			return 1;
	}

	return 0;	
}

void
show(char *cmd,char *topic,int ok,int nok)
{
	char *r,*rr;
	int c;

	r = dict_command(cmd);

	c = strtol(r,&rr,10);
	if(c == nok) {
		char tmp[255];
		sprint(tmp,"find any %s",topic);
		could_not(tmp,r);
	}
	if(rr == nil || c != ok) 
		could_not("get sane response",r);
	c = strtol(rr,nil,10);
	if(c <= 0)
		could_not("happen",r);

	print("%d %s found:\n",c,topic);

	do {
		char *r = dict_line();
		if(strcmp(".",r) == 0) {
			return;
		}
		print("%s\n",r);
	} while(c--);

	if(strncmp("250",dict_line(),3) != 0) //swallow 250
		could_not("find end of response. Protocol error",r);
}

int 
match(char *what)
{
	char cmd[1024];
	char *r,*rr;
	int c;

	if(need_quote(what))
		snprint(cmd,sizeof(cmd),"MATCH %s %s \"%s\"",dictdb,strategy,what);
	else
		snprint(cmd,sizeof(cmd),"MATCH %s %s %s",dictdb,strategy,what);	
	r = dict_command(cmd);
	c = strtol(r,&rr,10);
	if(c == 552) {
		fprint(2,"No matches for %s.\n",what);
		return 0;
	}
	if(rr == nil || c != 152)
		could_not("find any matches.",r);

	c = strtol(rr,nil,10);
	if(c <= 0)
		could_not("happen",r);
	print("%d matches found:\n",c);
	do {
		char *r = dict_line();

		if(strcmp(".",r) == 0) 
			break;
		print("%s -d %s\n",progname,r);
	} while(c--);

	if(strncmp("250",dict_line(),3) != 0) //swallow 250
		could_not("find end of response. Protocol error",r);
	return 1;
}

int 
define(char *what)
{
	char cmd[1024]; //should do, even for welsh names ?
	char *r,*rr;
	int c;

	if(need_quote(what))
		snprint(cmd,sizeof(cmd),"DEFINE %s \"%s\"",dictdb,what);
	else
		snprint(cmd,sizeof(cmd),"DEFINE %s %s",dictdb,what);	
	r = dict_command(cmd);
	c = strtol(r,&rr,10);
	if(c == 552) {
		fprint(2,"No definitions for %s.\n",what);
		return 0;
	}
	if(rr == nil || c != 150)
		could_not("find any definitions.",r);

	c = strtol(rr,nil,10);
	if(c <= 0)
		could_not("happen",r);
	print("%d definitions found:\n",c);
	read_definitions(c,what);
	
	return 1;
}

void
main(int argc, char *argv[])
{

	int needarg = 1;
	int cmd = Define;
	char *where = "tcp!dict.org!2628";
	progname = argv[0];

	ARGBEGIN {
	case 'w':
        	where = ARGF();
		break;
	case 'd':
		if((dictdb = ARGF()) == nil)
			usage();
		break;
	case 's':
		if((strategy = ARGF()) == nil)
			usage();
		break;
	case 'l':
		cmd = ShowDB;
		needarg = 0;
		break;
	case 't':
		cmd = ShowStrats;
		needarg = 0;
		break;
	case 'm':
		cmd = Match;
		break;
	default:
		print("Unknown flag '%c'\n", ARGC());
		break;
	} ARGEND

	if(argc < 1 && needarg) {
		usage();
		exits("unemployed");
	}
	
	fd = dial(where,nil,nil,nil);
	if(fd == -1) {
		perror("dial");
		exits("Could not connect");
	}

	Binits(&in,fd,OREAD,ibuf,sizeof(ibuf));
	Binit(&out,fd,OWRITE);
	shake_hands();
	
	switch(cmd) {
		case Define:
			for(;argc-- > 0 ; argv++)
				if(define(argv[0]) == 0)
					match(argv[0]);
			break;
		case Match:
			for(;argc-- > 0 ; argv++)
				match(argv[0]);
			break;
		case ShowDB:
			show("SHOW DB","database",110,554);
			break;
		case ShowStrats:
			show("SHOW STRAT","strategies",111,555);
			break;
		default:
			exits("impossible");
	}

	close(fd);
	exits(nil);	
}

void usage(void)
{
	print("Usage: define [-l|-t] [-m [-s strategy]] [-d db] word\n");
	exits("Bad arguments");
}
